from importlib import metadata
import sys

from aiohttp import ClientResponseError
import click


try:
    __version__ = metadata.version(__package__) if __package__ else "0.0.0"
except metadata.PackageNotFoundError:
    __version__ = "0.0.0"


def main() -> None:
    from zet.cli import zet

    try:
        zet()
    except ClientResponseError as ex:
        click.secho(f"Server returned HTTP {ex.status} {ex.message}", fg="red")
        sys.exit(1)



import json
from dataclasses import dataclass

from aiohttp import ClientResponse, ClientSession
from multidict import CIMultiDictProxy

from zet.entities import IncomingTrip, News, Route, Stop, Trip, TripStop, Vehicle


@dataclass(frozen=True)
class Response[T]:
    body: str
    status: int
    headers: CIMultiDictProxy[str]

    def json(self) -> T:
        return json.loads(self.body)

    @classmethod
    async def from_client_response(cls, response: ClientResponse):
        return cls(
            body=await response.text(),
            status=response.status,
            headers=response.headers,
        )


async def get_newsfeed(session: ClientSession) -> Response[list[News]]:
    async with session.get("/NewsProxyService.Api/api/newsfeed") as response:
        response.raise_for_status()
        return await Response.from_client_response(response)


async def get_routes(session: ClientSession) -> Response[list[Route]]:
    async with session.get("/TimetableService.Api/api/gtfs/routes") as response:
        response.raise_for_status()
        return await Response.from_client_response(response)


async def get_stops(session: ClientSession) -> Response[list[Stop]]:
    async with session.get("/TimetableService.Api/api/gtfs/stops") as response:
        response.raise_for_status()
        return await Response.from_client_response(response)


async def get_route_trips(session: ClientSession, route_id: str) -> Response[list[Trip]]:
    params = {"routeId": route_id, "daysFromToday": "0"}
    async with session.get("/TimetableService.Api/api/gtfs/routeTrips", params=params) as response:
        response.raise_for_status()
        return await Response.from_client_response(response)


async def get_vehicles(session: ClientSession) -> Response[list[Vehicle]]:
    async with session.get("/TransportService.Api/api/Vehicle") as response:
        response.raise_for_status()
        return await Response.from_client_response(response)


async def get_incoming_trips(session: ClientSession, stop_id: str) -> Response[list[IncomingTrip]]:
    path = "/TimetableService.Api/api/gtfs/stopIncomingTrips"
    params = {"stopId": stop_id, "isMapView": "false"}
    async with session.get(path, params=params) as response:
        return await Response.from_client_response(response)


async def get_trip_stop_times(session: ClientSession, trip_id: str) -> Response[list[TripStop]]:
    path = "/TimetableService.Api/api/gtfs/tripStopTimes"
    params = {"tripId": trip_id, "daysFromToday": 0}
    async with session.get(path, params=params) as response:
        return await Response.from_client_response(response)



import asyncio
from contextlib import asynccontextmanager
import logging
from types import SimpleNamespace
from aiohttp import ClientSession, TraceConfig, TraceRequestEndParams, TraceRequestStartParams

logger = logging.getLogger(__name__)


@asynccontextmanager
async def make_session():
    headers = {
        "appuid": "ZET.Mobile",
        "x-tenant": "KingICT_ZET_Public",
        "User-Agent": "okhttp/4.9.2",
    }

    async with ClientSession(
        base_url="https://api.zet.hr",
        headers=headers,
        trace_configs=[logger_trace_config()],
    ) as session:
        yield session


def logger_trace_config() -> TraceConfig:
    async def on_request_start(
        _: ClientSession,
        context: SimpleNamespace,
        params: TraceRequestStartParams,
    ):
        context.start = asyncio.get_event_loop().time()
        logger.debug(f"--> {params.method} {params.url}")

    async def on_request_end(
        _: ClientSession, context: SimpleNamespace, params: TraceRequestEndParams
    ):
        elapsed = round(100 * (asyncio.get_event_loop().time() - context.start))
        logger.debug(f"<-- {params.method} {params.url} HTTP {params.response.status} {elapsed}ms")

    trace_config = TraceConfig()
    trace_config.on_request_start.append(on_request_start)
    trace_config.on_request_end.append(on_request_end)
    return trace_config



import asyncio
from collections.abc import Coroutine
from functools import wraps
from typing import Callable, Concatenate, ParamSpec, TypeVar

from aiohttp import ClientSession
from zet.http import make_session

P = ParamSpec("P")  # Function parameters
R = TypeVar("R")  # Function return value


def async_command(f: Callable[P, Coroutine[None, None, R]]) -> Callable[P, R]:
    """
    Integrating click with asyncio:
    https://github.com/pallets/click/issues/85#issuecomment-503464628
    """

    @wraps(f)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
        c = f(*args, **kwargs)
        return asyncio.run(c)

    return wrapped


def pass_session(
    f: Callable[Concatenate[ClientSession, P], Coroutine[None, None, R]],
) -> Callable[P, Coroutine[None, None, R]]:
    """
    Pass the http session as first parameter.
    """

    @wraps(f)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        async with make_session() as session:
            return await f(session, *args, **kwargs)

    return wrapper



import logging
import textwrap

from aiohttp import ClientSession
import click

from zet import api, __version__
from zet.app import ZetApp
from zet.decorators import async_command, pass_session
from zet.entities import Stop
from zet.output import format_direction, format_route_type, format_time, format_tracked, table_dump


json_option = click.option(
    "--json",
    is_flag=True,
    help="Dump data as raw JSON instead of a human readable format.",
)


@click.group()
@click.option("--debug", is_flag=True, help="Enable debug logging")
@click.version_option(__version__)
def zet(debug: bool):
    if debug:
        logging.basicConfig(level=logging.DEBUG)


@zet.command()
@json_option
@async_command
@pass_session
async def news(session, json: bool):
    """Show news feed"""
    response = await api.get_newsfeed(session)

    if json:
        click.echo(response.body)
        return

    news = response.json()

    def _generator():
        first = True
        for item in news:
            if not first:
                yield click.style("-" * 80, dim=True) + "\n\n"
            yield click.style(item["title"], bold=True) + "\n\n"
            for line in textwrap.wrap(item["description"], 80):
                yield line + "\n"
            yield "\n"
            yield click.style(item["link"], underline=True, dim=True) + "\n"
            yield click.style(item["datePublished"], dim=True) + "\n\n"
            first = False

    click.echo_via_pager(_generator())


@zet.command()
@async_command
@pass_session
async def tui(session):
    app = ZetApp(session)
    await app.run_async()


@zet.command()
@click.option("-p", "--pager/--no-pager")
@json_option
@async_command
@pass_session
async def routes(session, pager: bool, json: bool):
    """List routes"""
    response = await api.get_routes(session)

    if json:
        click.echo(response.body)
        return

    routes = response.json()

    table_dump(
        routes,
        {
            "Type": lambda r: format_route_type(r["routeType"]),
            "No": lambda r: str(r["id"]),
            "From": lambda r: r["departureHeadsign"],
            "To": lambda r: r["destinationHeadsign"],
        },
        pager=pager,
    )


@zet.command()
@click.option("-p", "--pager/--no-pager")
@click.option("-n", "--name", help="Station name")
@json_option
@async_command
@pass_session
async def stops(session, pager: bool, name: str | None, json: bool):
    """List stops"""
    response = await api.get_stops(session)

    if json:
        click.echo(response.body)
        return

    stops = response.json()

    if name:
        stops = [s for s in stops if name in s["normalizedSearchName"]]

    table_dump(
        stops,
        {
            "ID": lambda s: str(s["id"]),
            "Name": lambda s: s["name"],
            "Latitude": lambda s: str(s["stopLat"]),
            "Longitude": lambda s: str(s["stopLong"]),
            "Type": lambda s: format_route_type(s["routeType"]),
            "Trips": lambda s: _trips(s),
        },
        pager=pager,
    )


def _trips(stop: Stop):
    return ", ".join(trip["routeCode"] for trip in stop["trips"])


@zet.command()
@click.argument("stop_id")
@click.option("-p", "--pager/--no-pager")
@click.option("-r", "--route", "route_id")
@json_option
@async_command
@pass_session
async def trips(session, stop_id: str, pager: bool, route_id: str, json: bool):
    """List arrivals for a given stop"""
    response = await api.get_incoming_trips(session, stop_id)

    if json:
        click.echo(response.body)
        return

    trips = response.json()

    if route_id:
        trips = [t for t in trips if t["routeShortName"] == route_id]

    table_dump(
        trips,
        {
            "#": lambda t: t["routeShortName"],
            "Destination": lambda t: t["headsign"],
            "Arrival": lambda t: format_time(t["expectedArrivalDateTime"]),
            "Tracked?": lambda t: format_tracked(t["hasLiveTracking"]),
        },
        pager=pager,
    )


@zet.command()
@click.argument("trip_id")
@click.option("-p", "--pager/--no-pager")
@json_option
@async_command
@pass_session
async def trip_stops(session, trip_id: str, pager: bool, json: bool):
    """List stops for a given trip"""
    response = await api.get_trip_stop_times(session, trip_id)

    if json:
        click.echo(response.body)
        return

    trips = response.json()

    table_dump(
        trips,
        {
            "#": lambda t: str(t["id"]),
            "Stop": lambda t: t["stopName"],
            "Arrival": lambda t: format_time(t["expectedArrivalDateTime"]),
            "Arrived?": lambda t: str(t["isArrived"]),
            "Predict?": lambda t: str(t["isArrivedPrediction"]),
        },
        pager=pager,
    )


@zet.command()
@click.argument("route_id")
@click.option("-p", "--pager/--no-pager")
@click.option(
    "-d",
    "--direction",
    type=click.Choice(["A", "B"], case_sensitive=False),
    help="Show only trips in the given direction",
)
@json_option
@async_command
@pass_session
async def route_trips(session, route_id: str, pager: bool, direction: str | None, json: bool):
    """List trips for a given route"""
    response = await api.get_route_trips(session, route_id)

    if json:
        click.echo(response.body)
        return

    trips = response.json()

    if direction == "A":
        trips = [t for t in trips if t["direction"] == 0]

    if direction == "B":
        trips = [t for t in trips if t["direction"] == 1]

    table_dump(
        trips,
        {
            "Trip ID": lambda t: str(t["id"]),
            "Direction": lambda t: format_direction(t["direction"]),
            "Headsign": lambda t: t["headsign"],
            "Depart": lambda t: format_time(t["departureDateTime"]),
            "Arrive": lambda t: format_time(t["arrivalDateTime"]),
            "Tracked?": lambda t: str(t["hasLiveTracking"]),
        },
        pager=pager,
    )


@zet.command()
@click.option("-p", "--pager/--no-pager")
@json_option
@async_command
@pass_session
async def vehicles(session: ClientSession, pager: bool, json: bool):
    """List vehicles"""
    response = await api.get_vehicles(session)

    if json:
        click.echo(response.body)
        return

    vehicles = response.json()

    table_dump(
        vehicles,
        {
            "#": lambda x: str(x["id"]),
            "Garage#": lambda x: x["garageNumber"],
            "Plate": lambda x: x["numberPlate"] or "",
            "Description": lambda x: x["description"],
        },
        pager=pager,
    )



from __future__ import annotations

from enum import IntEnum
from typing import TypedDict

type RouteId = int
type ShapeId = str
type StopId = int
type TripId = int
type VehicleId = int
type Direction = int


# GTFS Route Type
# https://gtfs.org/documentation/schedule/reference/#routestxt
class RouteType(IntEnum):
    TRAM = 0
    SUBWAY = 1
    RAIL = 2
    BUS = 3
    FERRY = 4
    CABLE = 5
    AERIAL = 6
    FUNICULAR = 7
    TROLLEYBUS = 11
    MONORAIL = 12


class News(TypedDict):
    title: str
    description: str
    link: str
    datePublished: str
    type: RouteType
    lines: list[int]
    stations: list[int]
    validFrom: str
    validTo: str


class Route(TypedDict):
    id: RouteId
    shortName: str
    longName: str
    routeType: RouteType
    departureHeadsign: str
    destinationHeadsign: str
    normalizedSearchName: str


class Stop(TypedDict):
    id: StopId
    name: str
    routeType: RouteType
    trips: list[StopTrip]
    stopLat: float
    stopLong: float
    parentStopId: StopId
    normalizedSearchName: str


class StopTrip(TypedDict):
    routeCode: str
    tripHeadsigns: list[str]


class Trip(TypedDict):
    id: TripId
    direction: Direction
    headsign: str
    departureDateTime: str
    arrivalDateTime: str
    hasLiveTracking: bool
    tripStatus: int  # ???
    vehicles: list[TripVehicle]
    shapeId: ShapeId


class TripVehicle(TypedDict):
    id: str
    position: Position


class TripStop(TypedDict):
    id: StopId
    stopName: str
    stopSequence: int
    expectedArrivalDateTime: str
    isArrived: bool
    isArrivedPrediction: bool
    stopLat: float
    stopLong: float
    trip: Trip


class Position(TypedDict):
    latitude: float
    longitude: float


class IncomingTrip(TypedDict):
    tripId: str
    routeShortName: str
    headsign: str
    expectedArrivalDateTime: str
    hasLiveTracking: bool
    daysFromToday: int


class Vehicle(TypedDict):
    id: VehicleId
    active: bool
    dateCreated: str
    garageNumber: str
    numberPlate: str | None
    vehicleTypeID: int
    isForDisabledPeople: bool
    isAutomaticImport: bool
    description: str
    factoryNumber: str



import asyncio
import textwrap
from aiohttp import ClientSession
from dateutil.parser import parse
from textual import work
from textual.app import App
from textual.containers import VerticalScroll
from textual.screen import Screen
from textual.widgets import DataTable, Footer, Header, Markdown, Static, TabPane, TabbedContent

from zet import api
from zet.entities import News, Route, Stop


class ZetApp(App[None]):
    def __init__(self, session: ClientSession):
        super().__init__()
        self.session = session
        self.animation_level = "none"

    async def on_mount(self):
        self.push_screen(LoadingScreen())
        self.load()

    @work
    async def load(self):
        stops, routes, news = await asyncio.gather(
            api.get_stops(self.session),
            api.get_routes(self.session),
            api.get_newsfeed(self.session),
        )
        self.switch_screen(MainScreen(stops, routes, news))


class MainScreen(Screen):
    def __init__(
        self,
        stops: list[Stop],
        routes: list[Route],
        news: list[News],
    ) -> None:
        super().__init__()
        self.stops = stops
        self.routes = routes
        self.news = news

    def compose(self):
        with TabbedContent():
            yield StopsPane(self.stops)
            yield NewsPane(self.news)


class StopsPane(TabPane):
    DEFAULT_CSS = """
    StopsPane {
        height: auto;

        DataTable {
            height: auto;
        }
    }
    """

    def __init__(self, stops: list[Stop]) -> None:
        super().__init__("Stops")
        self.stops = stops

    def compose(self):
        yield DataTable()

    def on_mount(self):
        dt = self.query_one(DataTable)
        dt.add_columns("ID", "Name", "SearchName")
        for stop in self.stops:
            dt.add_row(
                stop["id"],
                stop["name"],
                stop["normalizedSearchName"],
            )


class NewsPane(TabPane):
    def __init__(self, news: list[News]) -> None:
        super().__init__("News")
        self.news = news

    def compose(self):
        markdown = "\n".join(self._news_generator())
        with VerticalScroll():
            yield Markdown(markdown)

    def _news_generator(self):
        first = True
        for item in self.news:
            if not first:
                yield ""
                yield "---"
                yield ""
            yield f"**{item['title']}**"
            yield ""
            yield item["description"]
            yield ""
            yield item["link"]
            yield ""
            yield parse(item["datePublished"]).strftime("%d.%m.%Y %H:%M")
            first = False


class LoadingScreen(Screen):
    TITLE = "Loading"

    def compose(self):
        yield Header()
        yield Static("Loading...")
        yield Footer()



from typing import Callable, Iterable, List

import click
from dateutil.parser import parse

from zet.entities import Direction, RouteType

type TableSpec[T] = dict[str, Callable[[T], str]]


def table_dump[T](items: Iterable[T], spec: TableSpec[T], *, pager: bool = False):
    headers = list(spec.keys())
    rows = [[lmbda(item) for lmbda in spec.values()] for item in items]
    if pager:
        click.echo_via_pager(generate_table(headers, rows))
    else:
        click.echo("".join(generate_table(headers, rows)))


def generate_table(headers: list[str], data: list[List[str]]):
    widths = [[len(str(cell)) for cell in row] for row in data + [headers]]
    widths = [max(width) for width in zip(*widths)]

    def get_line(row):
        line = ""
        for idx, cell in enumerate(row):
            width = widths[idx]
            line += str(cell).ljust(width)
            line += "  "
        return line + "\n"

    underlines = ["-" * width for width in widths]

    yield get_line(headers)
    yield get_line(underlines)

    for row in data:
        yield get_line(row)


def format_direction(direction: Direction) -> str:
    if direction == 0:
        return "A"
    if direction == 1:
        return "B"
    return str(direction)


def format_time(value: str) -> str:
    return parse(value).time().strftime("%H:%M")


def format_relative_time(value: str) -> str:
    return parse(value).time().strftime("%H:%M")


def format_tracked(value: bool) -> str:
    return click.style("Tracked", fg="green") if value else click.style("Nope", fg="red")


def format_route_type(type: RouteType) -> str:
    match type:
        case RouteType.TRAM:
            return "Tram"
        case RouteType.SUBWAY:
            return "Subway"
        case RouteType.RAIL:
            return "Rail"
        case RouteType.BUS:
            return "Bus"
        case RouteType.FERRY:
            return "Ferry"
        case RouteType.CABLE:
            return "Cable"
        case RouteType.AERIAL:
            return "Aerial"
        case RouteType.FUNICULAR:
            return "Funicular"
        case RouteType.TROLLEYBUS:
            return "Trolleybus"
        case RouteType.MONORAIL:
            return "Monorail"
        case _:
            return f"Unknown ({type})"
